package com.atlassian;

import com.atlassian.fisheye.spi.data.ChangesetData;
import com.atlassian.fisheye.spi.data.FileRevisionData;
import com.atlassian.fisheye.spi.data.FileRevisionKeyData;
import com.atlassian.fisheye.spi.data.PathInfoData;
import com.atlassian.fisheye.spi.data.RepositoryData;
import com.atlassian.fisheye.spi.data.TabularQueryResultData;
import com.atlassian.fisheye.spi.services.ContentService;
import com.atlassian.fisheye.spi.services.RepositoryService;
import com.atlassian.fisheye.spi.services.RevisionDataService;
import com.atlassian.fisheye.spi.services.SearchService;
import org.springframework.beans.factory.annotation.Autowired;

import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.StringWriter;
import java.io.Writer;
import java.util.Arrays;
import java.util.Map;

public class ExampleServlet extends HttpServlet {
    private RepositoryService repositoryService;
    private RevisionDataService revisionDataService;
    private SearchService searchService;
    private ContentService contentService;

    @Autowired
    public ExampleServlet(RepositoryService repositoryService, RevisionDataService revisionDataService, SearchService searchService, ContentService contentService) {
        this.repositoryService = repositoryService;
        this.revisionDataService = revisionDataService;
        this.searchService = searchService;
        this.contentService = contentService;
    }

    public void doGet(HttpServletRequest request, HttpServletResponse response) throws IOException {

        PrintWriter writer = response.getWriter();

        printTop(writer);

        printRepositoryList(writer);

        String rep = getRunningRep();

        if (rep == null) {
            writer.println("\n\nNo accessible running repositories");
        } else {

            printPathList(rep, writer);

            printChangesetList(rep, writer);

            String changeset = revisionDataService.listChangesets(rep, "/", null, null, 1).getCsids().get(0);

            printChangeset(rep, changeset, writer);

            ChangesetData cs = revisionDataService.getChangeset(rep, changeset);
            FileRevisionKeyData rev = cs.getRevisions().get(0);

            printPathHistoryList(rep, rev.getPath(), writer);

            printRevision(rep, rev, writer);

            printQuery(rep, "select revisions from dir / where comment matches \"branch\" group by changeset", writer);

            printQueryAsRows(rep, "select revisions from dir / where comment matches \"branch\" group by changeset return path, revision, author, date, comment, csid, totalLines, linesAdded, linesRemoved, tags, reviews", writer);

            printContent(rep, rev, writer);

        }
        printBottom(writer);
        writer.flush();
        writer.close();
    }

    private void printContent(String rep, FileRevisionKeyData rev, PrintWriter writer) throws IOException {
        printHeader(writer, " contentService.getContent(" + rep + ", " + rev + ", response.getOutputStream()) ");
        writer.println("<blockquote>");
        writer.flush();

        File file = File.createTempFile("content", ".txt");
        OutputStream outputStream = new FileOutputStream(file);

        contentService.getContent(rep, rev, outputStream);
        outputStream.close();

        String text = escapeString(copyReaderToString(new FileReader(file)));
        writer.write(text);
        writer.flush();
        writer.println("</blockquote>");

    }


    public static String copyReaderToString(Reader src) throws IOException {
        StringWriter dest = new StringWriter();
        char[] buffer = new char[8 * 1024];
        int count = 0;
        do {
            dest.write(buffer, 0, count);
            count = src.read(buffer, 0, buffer.length);
        } while (count != -1);
        return dest.toString();
    }

    private void printQueryAsRows(String rep, String query, PrintWriter writer) {
        printHeader(writer, "searchService.queryAsRows(" + rep + ", " + query + ", 5)");

        printTableStart(writer);

        TabularQueryResultData result = searchService.queryAsRows(rep, query, 5);
        printTableRow(writer, result.getNames());

        for (Object[] row : result) {
            Object[] output = new Object[row.length];

            for (int i=0; i<row.length; i++) {
                switch (result.getColumnTypes()[i]) {
                    case STRING_ARRAY:
                        output[i] = Arrays.toString((String[])row[i]);
                        break;
                    case MAP_ARRAY:
                        output[i] = Arrays.toString((Map[]) row[i]);
                        break;
                    default:
                        output[i] = row[i];
                }
            }

            printTableRow(writer, output);
        }
        printTableEnd(writer);
    }

    private void printQuery(String rep, String query, PrintWriter writer) {
        printHeader(writer, "searchService.query(" + rep + ", " + query + ", 5)");

        writer.println("<ul>");
        for (FileRevisionKeyData rev : searchService.query(rep, query, 5)) {
            writer.println("<li> " + rev + " </li>");
        }
        writer.println("</ul>");
    }

    private void printRevision(String rep, FileRevisionKeyData rev, PrintWriter writer) {
        printHeader(writer, "revisionDataService.getRevision(" + rep + ", " + rev.getPath() + ", " + rev.getRev() + ")");

        writer.println("<table cellpadding=3>");

        FileRevisionData fileRev = revisionDataService.getRevision(rep, rev.getPath(), rev.getRev());

        printTableRow(writer, new Object[]{"path: ", fileRev.getPath()});

        printTableRow(writer, new Object[]{"rev: ", fileRev.getRev()});
        printTableRow(writer, new Object[]{"author: " , fileRev.getAuthor()});
        printTableRow(writer, new Object[]{"date: " , fileRev.getDate()});
        printTableRow(writer, new Object[]{"totalLines: " , fileRev.getTotalLines()});
        printTableRow(writer, new Object[]{"linesAdded: ", fileRev.getLinesAdded()});
        printTableRow(writer, new Object[]{"linesRemoved: ", fileRev.getLinesRemoved()});
        printTableRow(writer, new Object[]{"csid: ", fileRev.getCsid()});
        printTableRow(writer, new Object[]{"ancestor: ", fileRev.getAncestor()});
        printTableRow(writer, new Object[]{"state: ", fileRev.getState()});
        printTableRow(writer, new Object[]{"comment: ", fileRev.getComment()});

        printTableEnd(writer);
    }

    private void printPathHistoryList(String rep, String path, PrintWriter writer) {
        printHeader(writer, "revisionDataService.listPathHistory(" + rep + ", " + path + ");");

        writer.println("<table cellpadding=3>");
        for (FileRevisionData rev : revisionDataService.listPathHistory(rep, path)) {

            printTableRow(writer, new String[]{"path: " + rev.getPath(), "rev: " + rev.getRev(), "author: " + rev.getAuthor(), "date: " + rev.getDate(), "totalLines: " + rev.getTotalLines(),
                     "linesAdded: " + rev.getLinesAdded(), "linesRemoved: " + rev.getLinesRemoved(), "csid: " + rev.getCsid(),
                    "ancestor: " + rev.getAncestor(), "state: " + rev.getState(), "comment: " + rev.getComment()});
        }
        printTableEnd(writer);
    }

    private void printHeader(PrintWriter writer, String s) {
        writer.println("<h3>" + s + "</h3>");
    }

    private void printChangeset(String rep, String changeset, PrintWriter writer) {
        printHeader(writer, "revisionDataService.getChangeset(" + rep + ", " + changeset + ");");
        ChangesetData cs = revisionDataService.getChangeset(rep, changeset);

        writer.println("<table>");
        printTableRow(writer, new String[]{"Changeset id: ", cs.getCsid()});
        printTableRow(writer, new String[]{"Author : ", cs.getAuthor()});
        printTableRow(writer, new String[]{"Branch : ", cs.getBranch()});
        printTableRow(writer, new String[]{"Comment : ", cs.getComment()});
        printTableRow(writer, new Object[]{"Date : ", cs.getDate()});
        printTableRow(writer, new Object[]{"Revisions : ", cs.getRevisions()});

        printTableEnd(writer);
    }

    private void printChangesetList(String rep, PrintWriter writer) {

        printHeader(writer, "revisionDataService.listChangesets(" + rep + ", \"/\", null, null, 5);");

        writer.println("<ul>");
        for (String cs : revisionDataService.listChangesets(rep, "/", null, null, 5).getCsids()) {
            writer.println("<li> " + cs + " </li>");
        }
        writer.println("</ul>");
        writer.println("");
    }

    private void printPathList(String rep, PrintWriter writer) {
        printHeader(writer, "revisionDataService.listPaths(" + rep + ", \"/\");");


        printTableStart(writer);
        printTableRow(writer, new Object[]{"Name", "Is File", "Is Dir", "Is Head Deleted"});

        for (PathInfoData path : revisionDataService.listPaths(rep, "/")) {
            printTableRow(writer, new Object[]{path.getName(), path.isFile(), path.isDir(), path.isHeadDeleted()});
        }

        printTableEnd(writer);
    }

    private void printTableEnd(PrintWriter writer) {
        writer.print("</table>\n");
    }

    private String getRunningRep() {
        for (RepositoryData rep : repositoryService.listRepositories()) {
            if (RepositoryData.State.RUNNING.equals(rep.getState())) {
                return rep.getName();
            }
        }
        return null;
    }

    private void printBottom(PrintWriter writer) {
        writer.print("</body></html>");
    }

    private void printRepositoryList(PrintWriter writer) {
        printHeader(writer, "repositoryService.listRepositories()");

        printTableStart(writer);
        printTableRow(writer, new String[]{"Name", "State", "Enabled"});

        for (RepositoryData i : repositoryService.listRepositories()) {
            printTableRow(writer, new Object[]{i.getName(), i.getState(), i.isEnabled()});
        }
        printTableEnd(writer);
    }

    private void printTableStart(PrintWriter writer) {
        writer.print("<table cellpadding=\"2\" cellspacing=\"0\" border=\"1\">");
    }

    private void printTableRow(PrintWriter writer, Object[] elements) {
        writer.print("<tr>");
        for (Object element : elements) {
            writer.print("<td>" + element + "</td>");
        }
        writer.print("</tr>\n");
    }

    private void printTop(PrintWriter writer) {
        writer.print("<html>\n" +
                "  <head>\n" +
                "    <title> Sample fisheye plugin </title>\n" +
                "\n" +
                "  </head>\n" +
                "\n" +
                "<body>\n");
    }

    public static void escapeChar(char c, Writer w) throws IOException {
        if (c == '\r' || c == '\n') {
            w.write("<br>");
        } else if (c == '&')
            w.write("&amp;");
        else if (c == '<')
            w.write("&lt;");
        else if (c == '>')
            w.write("&gt;");
        else if (c == '"')
            w.write("&#034;");
        else if (c == '\'')
            w.write("&#039;");
        else if (((int) c) > 128) {
            w.write("&#" + ((int) c) + ";");
        } else {
            w.write(c);
        }
    }

    private static String escapeString(CharSequence s) {
        StringWriter out = new StringWriter();

        try {
            for (int i = 0; i < s.length(); i++) {
                char c = s.charAt(i);
                escapeChar(c, out);
            }
        } catch (IOException e) {
            // should never happen
        }
        return out.toString();
    }


}
